//
// Module: CirclePsfModel
// Author: Richard Saxton
//
// Description: 
//
// Sets up a PSF model for a circular source box.
// It is valid for a circular source box centred on the source.
//
// Modifications: See ChangeLog
// 
// $Id: CirclePsfModel.cc,v 1.2 2000/11/19 15:16:54 hsiddiqu Exp $

// General classes
#include <iostream>
#include <cmath>

// SAS-specific classes

// General utility functions
#include "errstr.h"
#include "Tracer.h"
#include "NumericalRange.h"
#include "ArfDataServer.h"
#include "EncircledEnergyDataServer.h"

// Local classes
#include "CirclePsfModel.h"

/* Constructor */
CirclePsfModel::CirclePsfModel( double Energy,    
		    double Theta,                
		    double Phi,                 
		    vector<double> Radii,             
		    unsigned int NGridEls     
		    ):
  _theta(Theta),
  _phi(Phi),
  _radii(Radii),
  _ngridels(NGridEls)
{
  PROC;

  /* Get the EE data server object */
   _eds = calServer.getAtom(_eds);

  /* Set up an initial encircled energy function */
   double emin = _eds->eValidity().lo();
   eef = new EncircledEnergyFunction(_eds->encircledEnergy(emin));

  /* Set the energy and set the PSF function*/
   energy(Energy);

   errstr 
    << "CirclePsfModel: constructed with "
    << " Energy " << _energy
    << " Theta " << Theta 
    << " Phi " << Phi 
    << " Radius " << Radii[0]
    << " NgridEls " << NGridEls
    << Sas::message(Msg::App, Msg::Verbose);


  ENDPROC;
};

CirclePsfModel::~CirclePsfModel()
{
  PROC;
  errstr 
    << "CirclePsfModel destroyed."
    << Sas::message(Msg::App, Msg::Verbose);

  ENDPROC;

}

/* Calculate integrated PSF over the circle or annulus. This works differently
   from other PSF implementations by calculating the flux within 
   a radius directly */

double CirclePsfModel::rad_psf() const
{
  PROC;

      errstr
        << "Getting Encircled energy (energy): "<<_energy
        << "radius "<<_radii[0]
        << "Ngridels "<<_ngridels
        << Sas::message(Msg::App, Msg::Verbose);

      double ee = (*eef)(_radii[0]) / _ngridels;

  /* Subtract the encircled energy from within the inner radius if
     this is an annulus */

      if (_radii.size() == 2) {
         ee = (*eef)(_radii[1]) / _ngridels - ee;
      }

  RETURN ee;
}

/* Method: double int_psf ( double xboxcentre,double yboxcentre,double xboxsize,
                     double yboxsize ) const
   Purpose: Integrate the PSF within a box
*/
double CirclePsfModel::int_psf( double xboxcentre, 
			  double yboxcentre, 
			  double xboxsize,  
			  double yboxsize ) const
{
  PROC;

   // Algorithm: Find the total PSF within the outer radius of 
   // the box, subtract the total PSF from within the inner radius
   // and normalise for the box size over the annulus area

   /* Set the source position centre as local variables */
      double srcxpos = arfDataServer->sourcePosition()->xdet(); 
      double srcypos = arfDataServer->sourcePosition()->ydet(); 
    
   /* Calculate the mean radius of this box */
      double boxRadius = sqrt( (srcxpos-xboxcentre) * (srcxpos-xboxcentre)
                        + (srcypos-yboxcentre) * (srcypos-yboxcentre) );
          
   /* This technique doesn't work for pixels right on the source
      centre so make sure radius is bigger than half a box */
      if (boxRadius < yboxsize/2.0+1) {
           boxRadius=yboxsize/2.0+1;
      }

   /* Find the PSF contribution in an annulus about this radius (in arcsec)  */
      double PsfCont = (*eef)( (boxRadius + yboxsize/2.0)*0.05 ) - 
                          (*eef)( (boxRadius - yboxsize/2.0)*0.05 );

   /* Normalise by area to get the PSF fraction in the box */
      double ee = PsfCont * yboxsize * xboxsize / ( M_PI * 
        ((boxRadius + yboxsize/2.0)*(boxRadius + yboxsize/2.0) -
         (boxRadius - yboxsize/2.0)*(boxRadius - yboxsize/2.0)) );

      errstr<<"Int_psf: src: "<<srcxpos<<" "<<srcypos
            <<"Centre: "<<xboxcentre<<" "<<yboxcentre
            <<" Widths: "<<xboxsize<<" "<<yboxsize
            <<"Radius: "<<boxRadius
            <<"psccont: "<<PsfCont
            <<"Normalised: "<<ee
            << Sas::message(Msg::App, Msg::Verbose);
 
  RETURN ee;
}

/* Mutator to reset the PSF energy */

void CirclePsfModel::energy(double Energy)
{
  PROC; 

  /* Ensure that the energy is not outside tabulated bounds
     Issue a warning message for high energy exceptions */ 

    if (Energy < _eds->eValidity().lo()) {
       _energy = _eds->eValidity().lo();
    } else if (Energy > _eds->eValidity().hi()){
       _energy = _eds->eValidity().hi()-0.001;
       errstr<<"Calculating PSF correction for energy "<<_energy
             <<Sas::warning("EEoutofBounds");
    } else {
       _energy = Energy;
    }

  /* Get the encircled energy as a function of radius */
    delete eef;
    eef = new EncircledEnergyFunction(_eds->encircledEnergy(_energy,
                              _theta, _phi));
    //eef = new EncircledEnergyFunction(_eds->encircledEnergy(_energy,
     //                         0.0,0.0));

  ENDPROC;
}

/* Mutator to reset the PSF circle radius */

void CirclePsfModel::setRadius(double Radius)
{
  PROC; 

        _radii[0] = Radius;

  ENDPROC;
}

void CirclePsfModel::setNgridEls(unsigned int NGridEls)
{
  PROC; 

   /* Sets the number of grid elements over which to calculate the PSF */
    _ngridels = NGridEls;

  ENDPROC;
}

double CirclePsfModel::psf(double /* x */,
                     double /* y */) const
{
  PROC; 

   /* Do nothing */
  RETURN 1.0;
}

string CirclePsfModel::mode() const
{
  PROC; 

  RETURN "CIRCLE";
}
