/// 
// Module: DetmapFilter
// Author: Hassan Siddiqui
//
// Description: 
//
// Given an input detmap, this class generates 
// a new PixelIterator for those pixels
// that are bounded by the specified region expression.
//
// Modifications: See ChangeLog
// 
// $Id: DetmapFilter.cc,v 1.8 2000/11/10 12:52:18 hsiddiqu Exp $

// General classes
#include <iostream>

// SAS-specific classes
#include "Dal.h"

// General utility functions
#include "errstr.h"
#include "Tracer.h"
#include "DataFilter.h"
#include "dss.h"

/* THIS CLASS CALLS attcalc, SO RUN AS METATASK */
#include <metatask.h>
#include "attcalc.h"

// Local classes
#include "DetmapFilter.h"
#include "TablePixelIterator.h"
#include "CompactTablePixelIterator.h"
#include "ArfDataServer.h"
#include "CheckExpr.h"

/* Constructor */
DetmapFilter::DetmapFilter():
  _init(false),
  //RDS _filteredset(0),
  set(0),
  _t(0)
{
  PROC;
  ENDPROC;
}


/* Destructor */
DetmapFilter::~DetmapFilter()
{
  PROC;
  errstr 
    << "Destroying DetmapFilter object .."
    << Sas::message(Msg::App, Msg::Verbose);

  if (_init)
    {
      delete _t; _t = 0;
      dataSetServer->close(_filteredset);
      //dataSetServer->close(set);
    }
  ENDPROC;
}

PixelIterator * const  DetmapFilter::filteredMap( PixelIterator * const detmap)
{
  PROC;
     /* If temporary filename is not given - get it from the filtered
        table name in the parameter list - RDS */
       RETURN filteredMap( detmap, arfDataServer->
                                        filteredTableName()+":EVENTS", false);
  ENDPROC;
}

PixelIterator * const  DetmapFilter::filteredMap( PixelIterator * const detmap,
const string & filteredtablename, bool insitu)
{
  PROC;

  if (_init) RETURN _t;
  
  /* Set up table name. Note that 'EVENTS' is used so that table
     can be processed by attcalc
  */
  string valuecolumnname = "VALUE";

  errstr<<" DetMapFilter called with tablename: "<<filteredtablename
   << Sas::message(Msg::App, Msg::Verbose);

  /* Create a table from the detmap, containing additional columns 
   * containing the DETX DETY coords (which can be used by attcalc)
   * and the sky coordinates (generated by attcalc) for each pixel */

  // RDS
     string rawtablename    = "rawpixellist.ds:EVENTS";    
  //if (! insitu) {
     
     set = dataSetServer->open(rawtablename, DataSet::Create);
  //} else {
  //rawtablename = filteredtablename;
  //if (arfDataServer->keepFilteredTable()) {
   //  set = dataSetServer->open(filteredtablename, DataSet::Create);
  //} 
  //else {
   //  set = dataSetServer->open(filteredtablename, DataSet::Temp);
  //}
  //}
  
  Table * table = set->addTable(rawtablename, detmap->size());
  // end - RDS
 
  errstr << "Table size is: "<<detmap->size()
         << Sas::message(Msg::App, Msg::Verbose);
  Column * valueCol  = table->addColumn(valuecolumnname, 
					Column::Real64, 
					"Value", 
					"");
  
  Column * skyxCol    = table->addColumn("X", 
					 Column::Int32, 
					 "Sky X", 
					 "");


  Column * skyyCol    = table->addColumn("Y", 
					 Column::Int32, 
					 "Sky Y", 
					 "");

  Column * detxCol    = table->addColumn("DETX", 
					 Column::Int16, 
					 "x det coords", 
					 "0.05 arcsec");


  Column * detyCol    = table->addColumn("DETY", 
					 Column::Int16, 
					 "y det coords", 
					 "0.05 arcsec");

  Column * indexxCol    = table->addColumn("XPIXEL", 
					 Column::Int16, 
					 "x index", 
					 "");

  Column * indexyCol    = table->addColumn("YPIXEL", 
					 Column::Int16, 
					 "y index", 
					 "");

  /* Have to add following entities for attcalc to work */
  Column * timeCol    = table->addColumn("TIME", 
					 Column::Real64, 
					 "elasped time", 
					 "sec");

  * set->addAttribute("TELESCOP",    "mission name") = 
    arfDataServer->telescope();
 
  * set->addAttribute("INSTRUME",    "instrument name") = 
    arfDataServer->instrument();
 
  * set->addAttribute("DATAMODE",    "instrument mode") = 
    arfDataServer->datamode();
 
  * set->addAttribute("DATE-OBS",    "Observation start date") = 
    arfDataServer->dateobs();
 
  * set->addAttribute("FILTER",      "Filter ID") = 
    arfDataServer->filter();
 
  /* NOTE Following are CCD-dependent attributes. 
   * using dummy values. 
   * Need to provide a better way of introducing
   * CCD info, perhaps by having separate event lists
   * for each CCD.
   */

  if (arfDataServer->instrument() == "EPN")
    {
      errstr 
	<< "DetmapFilter: Adding attributes to temp file appropriate to PN instrument "
	<< Sas::message(Msg::App, Msg::Verbose);
      /* Using 0070_0123700101/0070_0123700101_PNS00301IME.FIT (13 Oct 2000) */
      *set->addAttribute("CCDID",       "") = 0;
      *set->addAttribute("QUADRANT",    "") = 0;
      *set->addAttribute("CCDNODE",    "")  = 0;
      *set->addAttribute("DATATYPE",       "") = "IMAGE.EL";
      *set->addAttribute("WINDOWX0",    "")  = 0;
      *set->addAttribute("WINDOWY0",    "")  = 1;
      *set->addAttribute("WINDOWDX",    "")  = 64;
      *set->addAttribute("WINDOWDY",    "")  = 199;
      *set->addAttribute("EDUMODE",     "") = 3;
      *set->addAttribute("FRMTIME",     "") = 73;
      *set->addAttribute("CCDMODE",     "")  = 1
;
    }
  else
    {
      errstr 
	<< "DetmapFilter: Adding attributes to temp file appropriate to EMOS instrument "
	<< Sas::message(Msg::App, Msg::Verbose);

      /* Using 0070_0123700101/0070_0123700101_M1S00110IME.FIT (13 Oct 2000) */
      *set->addAttribute("CCDID",       "") = 1;
      *set->addAttribute("CCDNODE",    "")  = 0;
      *set->addAttribute("DATATYPE",       "") = "IMAGE.EL";
      *set->addAttribute("WINDOWX0",    "")  = 0;
      *set->addAttribute("WINDOWY0",    "")  = 0;
      *set->addAttribute("WINDOWDX",    "")  = 610;
      *set->addAttribute("WINDOWDY",    "")  = 602;
      *set->addAttribute("EDUMODE",       "") = 3;
      *set->addAttribute("FRMTIME",       "") = 26;
      *set->addAttribute("CCDNODE",    "")  = 0;

    }

  ColumnData * timeColData   = timeCol->data();
  real64     * timeData      = timeColData->real64Data();

  ColumnData * valueColData  = valueCol->data();
  real64     * valueData     = valueColData->real64Data();

  ColumnData * skyxColData   = skyxCol->data();
  int32      * skyxData      = skyxColData->int32Data();

  ColumnData * skyyColData   = skyyCol->data();
  int32      * skyyData      = skyyColData->int32Data();

  ColumnData * detxColData   = detxCol->data();
  int16      * detxData      = detxColData->int16Data();

  ColumnData * detyColData   = detyCol->data();
  int16      * detyData      = detyColData->int16Data();

  ColumnData * indexxColData   = indexxCol->data();
  int16      * indexxData      = indexxColData->int16Data();

  ColumnData * indexyColData   = indexyCol->data();
  int16      * indexyData      = indexyColData->int16Data();

  /* This conversion is HARDWIRED, but is probably acceptable
     as the DET coord definition is concrete. */
  double det2deg=1./(72000);

  /* Add WCS attributes to the DET columns (for attcalc) */
  WCSConvertor("DETX", 0, 0, det2deg, "DETY", 0, 0, det2deg).
                             addAttributesTo(detxCol, detyCol);

  /* Add WCS attributes to the index columns 
     (for pixel-DET coord mapping; this way the WCS info
     can also store the pixel size) */
  detmap->wcsconvertor()->addAttributesTo(indexxCol, indexyCol);

  /* Compute const time value to write to each row */
  double time = STime(arfDataServer->dateobs())-Xmm::missionReferenceTime;

  /* Now iterate through detmap, 
     adding an entry for each and every pixel */
  int row = 0;

  FOREACH(detmap)
    {
      PixelElement pixel = detmap->currentItem();
      timeData[row] = time;
      valueData[row]  = pixel.value();
      skyxData[row]   = 0;
      skyyData[row]   = 0;
      detxData[row]   = (int) pixel.xCentre();
      detyData[row]   = (int) pixel.yCentre();
      indexxData[row] = pixel.xIndex();
      indexyData[row] = pixel.yIndex();
      ++row;
    }

  /* Close dataset  */
  delete indexyColData;
  delete indexxColData;
  delete detyColData;
  delete detxColData;
  delete skyyColData;
  delete skyxColData;
  delete valueColData;
  delete timeColData;
  dataSetServer->close(set);

  /* Run attcalc if the DSS contains components
     defined in terms of the POS axes */
  const DataSubspace * dss = arfDataServer->dataSubspace();
  DSS::DSaxes axestest;
  axestest.push_back("POS");
  if (!CheckExpr::check(dss->selectExpression(axestest)))
      {
  
  /* Call ATTCALC to populate sky coordinate values */
  
  errstr << " Running attcalc on " << filteredtablename
	 << Sas::message(Msg::App, Msg::Verbose);

  /* ATTCALC changes the CAL accuracy mode (and who knows what else)
     so must save and restore this afterwards */
  AccuracyLevel tempmode = calServer.state()->accuracyLevel()->get();

  openParameters ("attcalc");
  setParameter   ("eventset",rawtablename);
  setParameter   ("attitudelabel","fixed");
  setParameter   ("fixedra", arfDataServer->rapnt());
  setParameter   ("fixeddec", arfDataServer->decpnt());
  setParameter   ("fixedposangle", arfDataServer->papnt());
  setParameter   ("refpointlabel","user");
  setParameter   ("nominalra", arfDataServer->raref());
  setParameter   ("nominaldec", arfDataServer->decref());
  setParameter   ("withatthkset", false);
  task(attcalc);
  closeParameters();
  openParameters("arfgen");

  /* Restore the CAL accuracy mode */
  calServer.state()->accuracyLevel()->set(tempmode);

      }
  else
    {
      errstr 
	<< "There does not seem to be any POS-related filters in the dss ["<<dss->selectExpression()
	<<". Will not run attcalc."
	<< Sas::message(Msg::App, Msg::Verbose);

    }

  /* If the insitu flag is set filter the exisiting table - if not
     use the command line parameters to create a new file */

  //RDS
    //string filteredtablename;

    //if (insitu) {
     //  _filteredset = set;

      // filteredtablename = filteredtablename.substr(0,filteredtablename.find(":"));
       //filteredtablename = "X"+filteredtablename;
	//  _filteredset = dataSetServer->clone( filteredtablename, 
	//			      filteredtablename, 
	//			      Dal::Modify);
    //} else {
       //dataSetServer->close(set);
       if (arfDataServer->keepFilteredTable())
	{
	  _filteredset = dataSetServer->clone( rawtablename, 
				      filteredtablename, 
				      Dal::Modify);
	}
       else
	{
	  _filteredset = dataSetServer->clone( rawtablename, 
				      filteredtablename,
				      Dal::Temp);

   /* Remove the raw table file */
          dataSetServer->remove(rawtablename);
	}
    //} 

   /* Now filter the dataset if required: filtering is Destructive */
    if (arfDataServer->filterDSS())
	{

	  /* Project region onto X,Y, DETX, DETY space */
	  DSS::DSaxes axes;
	  axes.push_back("POS");
	  axes.push_back("DET");
	  string projregionexpr = dss->selectExpression(axes);
      
	  errstr << "DetmapFilter: SLCTEXPR gives \'" 
		 << arfDataServer->selectionExpression() << "\'"
		 << Sas::message(Msg::App, Msg::Verbose);
	  
	  errstr << "DetmapFilter:  Filtering " << filteredtablename
		 << " using expression \'" << projregionexpr << "\'"
		 << Sas::message(Msg::App, Msg::Verbose);
	  
	  DataFilter filter(projregionexpr);
	  
         //RDS
	  table = _filteredset->table("EVENTS");
	  //table = set->table("EVENTS");
	  
	  long rows = filter.doFilter(*table);
	  
	  errstr << "DetmapFilter: got " << rows << " in filtered table."
		 << Sas::message(Msg::App, Msg::Verbose);

	}
      else
	{
	  errstr 
	    << "No filtering of pixel list has been performed."
	    << Sas::message(Msg::App, Msg::Verbose);
	}

      /* Pass filterSet to pixel iterator */
      if (insitu) {
 
       /* For the raw pixel iterator, everything is defined in INT16.
          Use the DETX column as a dummy for the valuecolumn */

         string nullcol;
         //RDS _t = new CompactTablePixelIterator( filteredtablename+":EVENTS",
         _t = new CompactTablePixelIterator( filteredtablename,
                  "XPIXEL", "YPIXEL", "DETX", nullcol);
      } else {
         //RDS
         _t = new TablePixelIterator( filteredtablename,
				"XPIXEL",
				"YPIXEL",
				valuecolumnname
				);
         _t->normalise();
      }

      _init=true;

      RETURN _t;
}


