// 
// Module: OffCCDPixelIterator
// Author: Richard Saxton 
//
// Description: 
//
// Class representation of an iterator that handles DAL table-based
//  detectormaps.
//
// Modifications: See ChangeLog
// 
// $Id: OffCCDPixelIterator.cc,v 1.2 2000/10/06 15:24:56 hsiddiqu Exp $

// General classes
#include <iostream>
#include <cmath>

// SAS-specific classes
#include "Dal.h"
#include "Region.h"
#include "ShapeScanner.h"
#include "CamCoords.h"
#include "GeometryDataServer.h"
#include "CalServer.h"
#include "Xmm.h"

// General utility functions
#include "errstr.h"
#include "Tracer.h"

// Local classes
#include "OffCCDPixelIterator.h"
#include "RegionExtent.h"
#include "PositionUtils.h"

/* Constructors */
OffCCDPixelIterator::OffCCDPixelIterator(const GeoRegion & sourceArea,
                                         const RegionExtent * extent,
                                         const PositionUtils * const putils,
                                         const bool ignoreEdgePixels) :
  _firstcalled(false),
  _offccdelements(0),
  _totalelements(0)
  
{
  PROC;
  
errstr<<"entered OffPixIt"
       << Sas::message(Msg::App, Msg::Noisy);

  using namespace scanningLib;

 /* Define the width of each grid element in detector coords 
    Use the nominal grid size of 1 arcsec for large areas (rad>2 amin)
    and 0.5 arcsecs for small (rad<2 amin) */

  _gridWidth=(extent->ymax()-extent->ymin())>4800 ? 20.0:10.0;

 /* Set up grid structure to evaluate which pixels lie within the
      source region. Work in detector coords. */
  struct Grid yGrid(_gridWidth, extent->ymin()+_gridWidth/2.0), 
              xGrid(_gridWidth, extent->xmin()+_gridWidth/2.0);
  //yGrid._origin = extent->ymin();
  //yGrid._stride = _gridWidth;
  //yGrid._length = UNT((extent->ymax()-extent->ymin()+1.0)/ _gridWidth);
  //xGrid._origin = extent->xmin();
  //xGrid._stride = _gridWidth;
  //xGrid._length = UNT((extent->xmax()-extent->xmin()+1.0)/ _gridWidth);

  errstr<<"Grid origin: "<<extent->ymin()
    <<" "<<UNT((extent->ymax()-extent->ymin()+1.0)/ _gridWidth)
    <<" "<<extent->xmin()
    <<" "<<UNT((extent->xmax()-extent->xmin()+1.0)/ _gridWidth)
    <<" stride: "<<_gridWidth
    <<" xmin: "<<xGrid.origin<<" ymin: "<<yGrid.origin
             << Sas::message(Msg::App, Msg::Verbose);
        // <<" length: "<<xGrid._length<<" "<<yGrid._length

 /* Define structures to hold region information */
  list<Span> spans;
  list<Span>::const_iterator iter;
  struct RegionScanner scanner( sourceArea, xGrid, yGrid );

 /* Loop over each grid element within these spans */
  int col=0;
  double detx, dety;
  int iy=0;
  CamCoords testpixel(0.0,0.0,0.0);
  PixCoords pixcoord(0.0,0.0);

  GeometryDataServer * ct = 0; ct = calServer.getAtom(ct);

 /* Get CCD size */

  unsigned minx = calServer.instrument()->minX();
  unsigned maxx = calServer.instrument()->maxX();
  unsigned miny = calServer.instrument()->minY();
  unsigned maxy = calServer.instrument()->maxY();

  errstr<<"Min/max is :"<<minx<<" "<<maxx<<" "<<miny<<" "<<maxy
             << Sas::message(Msg::App, Msg::Noisy);

 /* Loop over each column (X) */
  for (scanner.reset(); scanner.next(col, spans);) {

    detx = xGrid[col];

    for (iter = spans.begin(); iter !=spans.end(); ++iter) {

   /* Loop over each Y grid element within this span */
     for (iy=iter->min; iy<=iter->max; iy++) {

       _totalelements++;

       dety = yGrid[iy];
       testpixel = putils->detxy005ToCamCoord2(detx, dety);
       pixcoord  = putils->camCoord2ToPixCoord(testpixel);

       errstr<<"Test pixel is: "<<iy<<" "<<col<<" "<<" det: "<<detx<<" "<<dety
             <<"Raw: "<<pixcoord.x()<<" "<<pixcoord.y()
             << Sas::message(Msg::App, Msg::Noisy);

     /* Add off-chip pixels into off_ccd vector.
        If requested, also add edge-of-chip pixels into vector */

       if ( ct->onCcd(testpixel.x(), testpixel.y()) == Xmm::CCD_NONE ||
          ( ignoreEdgePixels && (pixcoord.x() == minx || pixcoord.x() == maxx
                         || pixcoord.y() == miny || pixcoord.y() == maxy) ) )
       {
        _OffCCDPixel.push_back(PixelElement(col,
                                            iy,
                                            detx, dety,
                                            _gridWidth, _gridWidth,
                                            1.0));
        _offccdelements++;

          errstr<<"OFF CCD: "<<testpixel.x()<<" "<<testpixel.y()
                <<" Det: "<<detx<<" "<<dety
                <<" Pixel: "<<pixcoord.x()<<" "<<pixcoord.y()
             << Sas::message(Msg::App, Msg::Noisy);

       }
       errstr<<"CCD: "<<ct->onCcd(testpixel.x(), testpixel.y())
           <<"Det: "<<detx<<" "<<dety
                <<" Pixel: "<<pixcoord.x()<<" "<<pixcoord.y()
           << Sas::message(Msg::App, Msg::Noisy);
    }
   }
  }

  /* Set up the WCS convertor to pass widths */
  _wcsconv = new WCSConvertor("DETX", 0, 0, _gridWidth,
                                 "DETY", 0, 0,_gridWidth);

  ENDPROC;
}

/* Destructor */
OffCCDPixelIterator::~OffCCDPixelIterator()
{
  PROC;
  ENDPROC;
}

/* Iterator methods */
void OffCCDPixelIterator::first()
{
  _firstcalled = true;
  _index=0;
}


void OffCCDPixelIterator::next ()
{
  if (isDone())
    {
      errstr<<"Anomaly in OffCCDPixelIterator - raise SPR"
          << Sas::error("ARFObjError");
    }

  _index++;
}

bool OffCCDPixelIterator::isDone()
{
  PROC;
    if (!_firstcalled)
    {
      errstr<<"Anomaly in OffCCDPixelIterator - raise SPR"
          << Sas::error("ARFObjError");
    }

  RETURN (_index >= _offccdelements);
}


PixelElement OffCCDPixelIterator::currentItem()
{
  PROC;

    if (!_firstcalled)
    {
      errstr<<"Anomaly in OffCCDPixelIterator - raise SPR"
          << Sas::error("ARFObjError");
    }

  RETURN _OffCCDPixel[_index];
}

unsigned OffCCDPixelIterator::size() const
{
  PROC;
  RETURN _offccdelements;
}

unsigned OffCCDPixelIterator::xbins() const
{
  PROC;
  
  RETURN _offccdelements;
}

unsigned OffCCDPixelIterator::totalPixels() const
{
  PROC;
  
  RETURN UNT(_totalelements * _gridWidth * _gridWidth);
}

/* WCSConvertor object */
const WCSConvertor * OffCCDPixelIterator::wcsconvertor() const
{
  PROC;
  RETURN _wcsconv;
}


