// 
// Module: PsfmodelPixelIterator
// Author: Hassan Siddiqui
//
// Description: 
//
// A detector map representation of a model PSF.
//
// Modifications: See ChangeLog
// 
// $Id: PsfmodelPixelIterator.cc,v 1.8 2000/12/19 14:00:15 hsiddiqu Exp $

/* General classes */

/* SAS-specific classes */
#include "Dal.h"
#include "errstr.h"
#include "Tracer.h"
#include "ArfDataServer.h"

/* Local classes */
#include "PsfmodelPixelIterator.h"

/* Constructors */

PsfmodelPixelIterator::
PsfmodelPixelIterator( double xmin,
		       double xmax,
		       unsigned xbins,
		       double ymin, 
		       double ymax,
		       unsigned ybins):
  _firstcalled(false),
  _xbins(xbins),
  _wcsconv(0)
{ 
  PROC;
  
  errstr 
    << "Constructing a PSF detmap with "
    << " xmin="  << xmin
    << " xmax="  << xmax
    << " xbins=" << xbins
    << " ymin="  << ymin
    << " ymax="  << ymax
    << " ybins=" << ybins
    << Sas::message(Msg::App, Msg::Verbose);
 
  if (xmin >= xmax)
    {
      errstr << " X range for map ( min =" << xmin
	     << ", max ="<<xmax
	     << " ) is invalid. "
	     << Sas::error("MapXRangeInvalid");

      ALTRETURN;
    }

  if (ymin >= ymax)
    {
      errstr << " Y range for map ( min =" << ymin
	     << ", max ="<<ymax
	     << " ) is invalid. "
	     << Sas::error("MapYRangeInvalid");
    }


  _size = xbins * ybins;

  if (_size == 0)
    {
      errstr << " The number of bins requested is zero "
	     << " ( xbins =" << xbins
	     << ", ybins =" << ybins
	     << Sas::error("MapSizeInvalid");
    }

  double xdelt = (xmax-xmin)/xbins;
  double ydelt = (ymax-ymin)/ybins;


  /* Create WcsConvertor object */
  _wcsconv = new WCSConvertor(
			      "DETX",
			      1,  // 1st pixel in array is [1,1]
			      xmin+(xdelt/2), // bin centre
			      xdelt,
			      "DETY",
			      1, // 1st pixel in array is [1,1]
			      ymin+(ydelt/2),
			      ydelt);

  ENDPROC;
}


/* Destructor */
PsfmodelPixelIterator::~PsfmodelPixelIterator()
{
  PROC;
  delete _wcsconv;   _wcsconv=0;
  ENDPROC;
}


/* Iterator methods */
void PsfmodelPixelIterator::first()
{
  _firstcalled = true;
  _index=0;
  _xindex=1;
  _yindex=1;
}

void PsfmodelPixelIterator::next()
{
  PROC;
  if (isDone())
    {
      errstr<<"Anomaly in PsfmodelPixelIterator - raise SPR"
          << Sas::error("ARFObjError");
    }

  _index++;

  _xindex++;
  if (_xindex > _xbins)
    {
      _xindex=1;
      _yindex++;
    }

  RETURN;
}

bool PsfmodelPixelIterator::isDone()
{
  PROC;
    if (!_firstcalled)
    {
      errstr<<"Anomaly in PsfmodelPixelIterator - raise SPR"
          << Sas::error("ARFObjError");
    }
  
  errstr 
    << "PsfmodelPixelIterator::isDone() got "
    << " index = "  << _index
    << " size = "   << _size
    << " xindex = " << _xindex
    << " yindex = " << _yindex
    << Sas::message(Msg::App, Msg::Noisy);

  RETURN (_index>=_size);
}


PixelElement PsfmodelPixelIterator::currentItem()
{
  PROC;

  /* Convert x and y positions to detector coordinates, 
     using WCS attributes */
  CoordPair physxy = _wcsconv->physicalCoords(_xindex, _yindex);
  double x = physxy.x();
  double y = physxy.y();

  double value = arfDataServer->psf()->int_psf( x, y,
					 _wcsconv->xWidth(),
					 _wcsconv->yWidth() );

  PixelElement result (_xindex,
		       _yindex,
                       x,
                       y,
		       _wcsconv->xWidth(),
		       _wcsconv->yWidth(),
		       value);

  RETURN result;
}

unsigned PsfmodelPixelIterator::size() const
{
  PROC;
  RETURN _size;
}

unsigned PsfmodelPixelIterator::xbins() const
{
  PROC;
  RETURN _xbins;
}

/* WCSConvertor object */
const WCSConvertor * PsfmodelPixelIterator::wcsconvertor() const
{
  PROC;
  RETURN _wcsconv;
}

